Graphics3D 800,600,32,2
SetBuffer BackBuffer()


; Variables to change
Global Speed#=0.02			; Player move Speed
Global scale%=100			; Scene scale
Global angle#=160			; Initial light angle (0...90...180...270...360 = dawn,noon,evening,midnight,dawn)
Global tilesize%=32			; Vertices per tile (x*x)
Global VertexColors%=False	; use Vertexcolors or not
Global Colored=True			; show tile seams (switch with Key "C")
Global Fog%=True			; use fog or not (switch with Key "F")
Global divider#=2.0			; vertical divider for height


; Globals
Global min#=65536.0
Global max#=0.0
Global oldangle#=-1
Global minm#,maxm#
Global size,xpts,ypts,xscale#,yscale#,zscale#,crad#,crvm,hscale,hbase
Global cam%,speedmulti%,wf%,init%
Global lightpivot%,sun%,light%,ambientcol%,clscol%,water%
Global FPS_C%,FMS%,FPS%
Global heightfield%,tiles%,tsize%
Global Tween#


; Arrays
Dim meshes(0,0),surfaces(0,0),tris(0,0),vertex(0,0)


; Loading Textures
Global colormap=LoadTexture("colormap.jpg",16+32)
Global texsize=TextureWidth(colormap)
ScaleTexture colormap,texsize,texsize


; Load Terragen Terrain
Print "Loading Terrain..." : LoadTerragen("island.ter")


; Init Scene
Print "Init Scene..." : InitScene()


; Create Meshterrain Tiles
Print "Create Meshterrain Tiles..." : CreateTiles()


; Fixing Chunk normals
Print "Fixing Chunk Normals..." : FixChunkNormals()


; Collisions
Collisions 1,2,2,3
MoveMouse GraphicsWidth()/2,GraphicsHeight()/2


While Not KeyHit(1)
	
	Local ms%,ft%
	
	; Frame Tweeing
	ms=MilliSecs() : Tween#=(ms-ft)/5.0 : ft=ms
	
	; check for User Input
	CheckInput()
	
	Updatescene()
	
	; Render
	UpdateWorld Tween
	RenderWorld
	
	; Infotext
	Text 0, 0,"Tris..........: "+TrisRendered()
	Text 0,15,"FPS...........: "+ShowFPS()
	Text 0,30,"Angle.........: "+Int(angle)+""
	Text 0,45,"Tiles/Tilesize: "+(tiles*tiles)+"/"+tsize
	
	Flip 0
	
Wend

End


; -------------------------------------------------------------------------------------------------
; simple FPS measurement
; -------------------------------------------------------------------------------------------------
Function ShowFPS()
	
    FPS_C=FPS_C+1 : If FMS<MilliSecs() Then FMS=MilliSecs()+1000 : FPS=FPS_C : FPS_C=0
	Return FPS
	
End Function


; -------------------------------------------------------------------------------------------------
; Check User Input (Keyboard / Mouse) and Camera Movement
; -------------------------------------------------------------------------------------------------
Function CheckInput()
	
	Local up%=0
	
	; Key Q/W: change light angle
	If KeyDown(16) Then angle=angle-0.1*Tween
	If KeyDown(17) Then angle=angle+0.1*Tween
	
	; Key 1-9: predefined angle 0-180 in 22.5 steps
	If KeyHit(2) Then angle=0.0
	If KeyHit(3) Then angle=22.5
	If KeyHit(4) Then angle=45
	If KeyHit(5) Then angle=67.5
	If KeyHit(6) Then angle=90
	If KeyHit(7) Then angle=112.5
	If KeyHit(8) Then angle=135
	If KeyHit(9) Then angle=157.5
	If KeyHit(10) Then angle=180.0
	
	; SHIFT = increase player speed 5 times
	If KeyDown(54) Then speedmulti=5 Else speedmulti=1
	
	; Space = Wireframe on/off
	If KeyHit(57) Then wf=1-wf : WireFrame wf
	
	; F = Fogswitch
	If KeyHit(33) Then Fog=1-Fog : CameraFogMode cam,Fog
	
	; C = Colorswitch
	If KeyHit(46) Then Colored=1-Colored : FixChunkNormals()
	
	; LMB/RMB: up/down
	If MouseDown(1) Then up=1
	If MouseDown(2) Then up=-1
	
	; Simple Camera movement
	TurnEntity cam,MouseYSpeed(),-MouseXSpeed(),0
	RotateEntity cam,EntityPitch(cam),EntityYaw(cam),0
	MoveEntity cam,(KeyDown(205)-KeyDown(203))*Speed*Tween*speedmulti/2.0,up*Speed*Tween*speedmulti,(KeyDown(200)-KeyDown(208))*Speed*Tween*speedmulti/2
	
	MoveMouse GraphicsWidth()/2,GraphicsHeight()/2
	
End Function


; -------------------------------------------------------------------------------------------------
; Init additional elements like cam, light, water
; -------------------------------------------------------------------------------------------------
Function InitScene()
	
	; Create Camera
	cam=CreateCamera()
	PositionEntity cam,scale/3,scale/6,scale/3
	CameraRange cam,0.01,scale*2
	CameraFogMode cam,Fog
	CameraFogRange cam,scale/2,scale
	EntityType cam,1
	EntityRadius cam,0.1
	CameraClsColor cam,187,213,239
	CameraFogColor cam,187,213,239
	
	; Light
	lightpivot=CreatePivot()
	sun=CreateSphere(8,lightpivot)
	ScaleEntity sun,2,2,2
	EntityColor sun,255,255,128
	EntityFX sun,1+8
	PositionEntity sun,-scale/2,0,0
	light=CreateLight(1,sun)
	PositionEntity light,0,0,0
	LightColor light,255,255,255
	AmbientLight 32,32,32
	
	; Water
	water=CreateCube()
	EntityFX water,1+8
	EntityAlpha water,0.5
	EntityColor water,32,64,128
	ScaleEntity water,scale/2,0.01,scale/2
	
End Function


; -------------------------------------------------------------------------------------------------
; Updates light FX by angle
; -------------------------------------------------------------------------------------------------
Function Updatescene()
	
	; Limit angle to 0...360 degrees
	angle=angle Mod 360.0 : If angle<0 Then angle=360.0+angle
	
	; only continue if angle has been changed
	If angle<>oldangle Then
		
		; rotate light angle and point light sources to pivot
		RotateEntity lightpivot,0,-90,-angle
		PointEntity light,lightpivot
		
		; Update Ambientlight, Cameraclscolor according to angle
		If angle>=0 And angle<=180 Then
			
			If angle<=90 Then
				ambientcol=Normalize(angle,0,90,32,128)
				clscol=Normalize(angle,0,90,16,512)
			EndIf
			
			If angle>=90 Then
				ambientcol=Normalize(angle,90,180,128,32)
				clscol=Normalize(angle,90,180,512,16)
			EndIf
			
			; limit daylight
			If clscol>224 Then clscol=224
			
			; set AmbientLight And CameraClsColor
			AmbientLight ambientcol,ambientcol,ambientcol
			CameraClsColor cam,clscol*0.70,clscol*0.75,clscol*1.0
			CameraFogColor cam,clscol*0.75,clscol*0.80,clscol*1.0
			
		EndIf
		
	EndIf
	
	If Not init Then PointEntity cam,lightpivot : init=True
	
	oldangle=angle
	
End Function


; -------------------------------------------------------------------------------------------------
; Creates and textures the tiles, optional with vertexcolors by the colormap texture
; -------------------------------------------------------------------------------------------------
Function CreateTiles()
	
	Local x%,y%
	
	LockBuffer TextureBuffer(colormap)
	
	For y=0 To tiles-1
		
		For x=0 To tiles-1
			
			; new Tile
			meshes(x,y)=CreateMeshTile(scale/tsize,tsize+1,x,y)
			EntityFX meshes(x,y),2
			
			; get Surface and Tris per Tile
			surfaces(x,y)=GetSurface(meshes(x,y),1)
			tris(x,y)=CountVertices(surfaces(x,y))
			
			; apply texture
			EntityTexture meshes(x,y),colormap,0,1
			
			; collision
			EntityType meshes(x,y),2
			
		Next
		
	Next
	
	UnlockBuffer TextureBuffer(colormap)
	
End Function


; -------------------------------------------------------------------------------------------------
; Fixes the Chunk normals that no visible seams appear
; -------------------------------------------------------------------------------------------------
Function FixChunkNormals()
	
	Local x%,y%,z%
	Local nx#,ny#,nz#
	Local r%,g%,b%
	
	For y=0 To tiles-1
		
		For x=0 To tiles-1
			
			For z=0 To tsize
				
				; align vertical normals
				nx=VertexNX(surfaces(x,y),tsize+(z*(tsize+1)))
				ny=VertexNY(surfaces(x,y),tsize+(z*(tsize+1)))
				nz=VertexNZ(surfaces(x,y),tsize+(z*(tsize+1)))
				If x<tiles-1 Then VertexNormal surfaces(x+1,y),0+(z*(tsize+1)),nx,ny,nz
				
				; align horizontal normals
				nx=VertexNX(surfaces(x,y),tris(x,y)-(tsize+1)+z)
				ny=VertexNY(surfaces(x,y),tris(x,y)-(tsize+1)+z)
				nz=VertexNZ(surfaces(x,y),tris(x,y)-(tsize+1)+z)
				If y<tiles-1 Then VertexNormal surfaces(x,y+1),z,nx,ny,nz
				
				r=128
				g=128
				b=128
				
				; If set: seams=red
				If Colored Then
					r=255
					g=0
					b=0
				EndIf
				
				VertexColor surfaces(x,y),tsize+(z*(tsize+1)),r,g,b
				If x<tiles-1 Then VertexColor surfaces(x+1,y),(z*(tsize+1)),r,g,b
				VertexColor surfaces(x,y),tris(x,y)-(tsize+1)+z,r,g,b
				If y<tiles-1 Then VertexColor surfaces(x,y+1),z,r,g,b
					
			Next
			
		Next
	Next
	
End Function


; -------------------------------------------------------------------------------------------------
; Creates a single Mesh Terrain Tile with unfixed chunk borders
; -------------------------------------------------------------------------------------------------
Function CreateMeshTile(size%,vertexes,startx%=0,starty%=0)
	
	Local a%,r%,g%,b%,rgb%,v%
	Local mesh%,surface%
	Local xpoint%,ypoint%
	Local zpoint%
	Local tri%,lx%,ly%,h#,index%
	Local texfactor1#=texsize/(xpts-1)
	Local texfactor2#=texsize/(xpts*1.0)
	Local vx#,vz#
	Local sx%=startx*tsize
	Local sy%=starty*tsize
	
	; create mesh and surface
	mesh = CreateMesh()
	surface = CreateSurface(mesh)
	
	Local buffer%=TextureBuffer(colormap)
	
	LockBuffer buffer
	
    ; set vertices
	For xpoint=0 To vertexes-1
		
		For zpoint=0 To vertexes-1
			
			h#=Normalize(PeekShort(heightfield,(((startx*tsize)+zpoint)+((starty*tsize)+xpoint)*xpts)*2)-32767,min,max,minm,maxm)/divider
			
			vx=-(scale/2.0)+(((starty)*tsize)+xpoint)*(scale*1.0/xpts)
			vz=-(scale/2.0)+(((startx)*tsize)+zpoint)*(scale*1.0/xpts)
			
			v=AddVertex (surface,vx,h,vz,((startx*tsize)+zpoint)*texfactor1,((starty*tsize)+xpoint)*texfactor1)
			
			If VertexColors Then
				
				; read texel
				rgb=ReadPixelFast(((startx*tsize)+zpoint)*texfactor2,((starty*tsize)+xpoint)*texfactor2,buffer)
				r=(rgb And $ff0000)/$10000
				g=(rgb And $ff00)/$100
				b=(rgb And $ff)
				
				; set vertexcolor by colormap
				VertexColor surface,v,r,g,b,1
				
			Else
				
				VertexColor surface,v,128,128,128,1.0
				
			EndIf	
			
			
		Next
		
	Next
	
	UnlockBuffer buffer
	
	; create triangles
	For tri = 0 To (vertexes^2)-vertexes-1
		If tri Mod vertexes<>vertexes-1
			AddTriangle (surface,tri,tri+1,tri+vertexes)
			AddTriangle (surface,tri+1,tri+vertexes+1,tri+vertexes)
		EndIf
	Next
	
	UpdateNormals mesh
	
	Return mesh
	
End Function


; -----------------------------------------------------------------------------
; Read Terragen-File and store heightdata to bank
; -----------------------------------------------------------------------------
Function LoadTerragen(filename$)
	
	Local bank%,i%,xpt%,ypt%,h%,h1#,bytes%=16,pos%,off%=0
	Local infoflag$,sizeflag$,xptsflag$,yptsflag$,scalflag$,cradflag$,crvmflag$,altwflag$,endflag$
	
	; open stream
	Local file=ReadFile(filename$)
	
	; Terragen-Header available?
	For i=1 To 16 : infoflag=infoflag+Chr(ReadByte(file)) : Next
	If infoflag<>"TERRAGENTERRAIN " Then RuntimeError "No valid Terragen Header found!"
	
	; SIZE-Flag available?
	For i=1 To 4 : sizeflag=sizeflag+Chr(ReadByte(file)) : Next
	If sizeflag<>"SIZE" Then RuntimeError "No SIZE Flag found!" Else size=ReadInt(file) : bytes=bytes+4+4
	
	; XPTS-Flag available?
	For i=1 To 4 : xptsflag=xptsflag+Chr(ReadByte(file)) : Next
	If xptsflag<>"XPTS" Then RuntimeError "No XPTS Flag found!" Else xpts=ReadInt(file) : bytes=bytes+4+4
	
	; YPTS-Flag available?
	For i=1 To 4 : yptsflag=yptsflag+Chr(ReadByte(file)) : Next
	If yptsflag<>"YPTS" Then RuntimeError "No YPTS Flag found!" Else ypts=ReadInt(file) : bytes=bytes+4+4
	
	; SCAL-Flag available?
	For i=1 To 4 : scalflag=scalflag+Chr(ReadByte(file)) : Next
	If scalflag<>"SCAL" Then RuntimeError "No SCAL Flag found!" Else xscale#=ReadFloat(file) : yscale#=ReadFloat(file) : zscale#=ReadFloat(file) : bytes=bytes+4+4+4+4
	
	; CRAD-Flag available? (optional)
	For i=1 To 4 : cradflag=cradflag+Chr(ReadByte(file)) : Next
	If cradflag<>"CRAD" Then SeekFile file,FilePos(file)-4 : Goto nextcrvm Else crad=ReadFloat(file) : bytes=bytes+4+4
	
	; CRVM-Flag available? (optional)
	.nextcrvm
	For i=1 To 4 : crvmflag=crvmflag+Chr(ReadByte(file)) : Next
	If crvmflag<>"CRVM" Then SeekFile file,FilePos(file)-4 : Goto nextaltw Else crvm=ReadFloat(file) : bytes=bytes+4+4
	
	; ALTW-Flag available?
	.nextaltw
	For i=1 To 4 : altwflag=altwflag+Chr(ReadByte(file)) : Next
	If altwflag<>"ALTW" Then RuntimeError "No ALTW Flag found!" Else hscale=ReadShort(file) : hbase=ReadShort(file) : bytes=bytes+4+2+2
	
	; Banksize = x*y*2
	heightfield=CreateBank(xpts*ypts*2)
	
	; read height data
	For ypt=ypts To 1 Step -1
		
		For xpt=1 To xpts Step 1
			
			; get 16bit height info and convert to signed integer
			h=ReadShort(file)-32767
			
			; calc minimum/maximum height
			If h<=min Then min=h
			If h>=max Then max=h
			
			; put height data to bank
			PokeShort heightfield,((xpt-1)+(ypt-1)*xpts)*2,h
			
		Next
	
	Next
	
	; close stream
	CloseFile(file)
	
	; calc number of tiles and tilesize
	tiles%=(xpts-1)/tilesize
	tsize%=(xpts-1)/tiles
	
	; get minimum/maximum height
	minm#=(hbase+((min*hscale)/65536.0))*scale/xpts*(xpts/1025.0)
	maxm#=(hbase+((max*hscale)/65536.0))*scale/xpts*(xpts/1025.0)
	
	Dim meshes(tiles,tiles),surfaces(tiles,tiles),tris(tiles,tiles),vertex(tilesize,tilesize)
	
End Function


; -----------------------------------------------------------------------------
; Normalize a value
; -----------------------------------------------------------------------------
Function Normalize#(value#=128.0,value_min#=0.0,value_max#=255.0,norm_min#=0.0,norm_max#=1.0)
	
	Return ((value#-value_min#)/(value_max#-value_min#))*(norm_max#-norm_min#)+norm_min#
	
End Function


; -------------------------------------------------------------------------------------------------
; Create a simple quad, with alpha and fx
; -------------------------------------------------------------------------------------------------
Function CreateQuad(r%=255,g%=255,b%=255,alpha#=1.0,fx%=0)
	
	Local mesh%,surface%,v1%,v2%,v3%,v4%
	
	mesh=CreateMesh()
	surface=CreateSurface(mesh)
	
	v1=AddVertex (surface,-1,1,0,1,0)
	v2=AddVertex (surface,1,1,0,0,0)
	v3=AddVertex (surface,-1,-1,0,1,1)
	v4=AddVertex (surface,1,-1,0,0,1)
	
	VertexColor surface,v1,r,g,b,alpha
	VertexColor surface,v3,r,g,b,alpha
	VertexColor surface,v2,r,g,b,alpha
	VertexColor surface,v4,r,g,b,alpha
	
	AddTriangle(surface,0,1,2)
	AddTriangle(surface,3,2,1)
	
	EntityFX mesh,fx
	
	FlipMesh mesh
	
	Return mesh
	
End Function
;~IDEal Editor Parameters:
;~C#Blitz3D